# -*- coding: utf-8 -*-
"""
Created on Wed Apr 12 15:00:08 2017

@author: Kevin Anderson
"""

import pandas as pd
import numpy as np
import datetime as datetime
import os

from ccrenew.dashboard.data_processing.CCR import file_project

def USB1_performance_guarantee(df, length, df_Pvsyst, PIS_date, guarantee_input, df_perf):
    if PIS_date.month != 12 or PIS_date.year != 2014:
        raise RuntimeError("Guarantee assumptions not met")
    n_years = df.index[-1].year - 2015
    
    # calculate how much each month contributes to yearly POA
    monthly_weights = df_Pvsyst['POA (W/m2)'].resample('M').sum() / df_Pvsyst['POA (W/m2)'].sum()
    # calculate what percentage of each month we have
    monthly_fractions = df['POA_avg'].resample("M").count() / df_Pvsyst['POA (W/m2)'].resample("M").count()
    monthly_fractions = monthly_fractions.fillna(0)

    proration = (monthly_weights * monthly_fractions).sum()
    # calculate degradation, prorated by monthly weights and time of year
    prorated_guarantee_input = guarantee_input * 1000.0 * proration
    degraded_guarantee_input = prorated_guarantee_input * (0.993 ** n_years)
    
    df_year = df.resample("A").sum()
    actual_production = df_year['Meter_Corrected_2'].item()
    
    df_perf.loc[df.index.date[-1], ['Guar_freq','kwh_guar','kwh_pro','usb_years','kwh_pro_deg','kwh_meas','USB1_target']] = ['Annual', guarantee_input, prorated_guarantee_input, n_years, degraded_guarantee_input, actual_production, 1]
    df_perf['Gaur_range'] = '1/1 - 12/31'
    
    if actual_production > degraded_guarantee_input:
        df_perf['USB_result'] = 1
        return df_perf, [1] * length + [np.nan]*(12-length)
    
    POA_adjustment = df_year['POA_avg'].item() / (df_Pvsyst['POA (W/m2)'].resample("M").sum() * monthly_fractions).sum()
    if actual_production > degraded_guarantee_input * POA_adjustment:
        df_perf[['kwh_meas_poa','USB_result']] = [degraded_guarantee_input * POA_adjustment, actual_production/degraded_guarantee_input / POA_adjustment]
        return df_perf, [1] * length + [np.nan]*(12-length)
    
    GHI_adjustment = df_year['GHI_avg'].item() / (df_Pvsyst['GHI (W/m2)'].resample("M").sum() * monthly_fractions).sum()
    if actual_production > degraded_guarantee_input * GHI_adjustment:
        df_perf[['kwh_meas_ghi','USB_result']] = [degraded_guarantee_input * GHI_adjustment, actual_production/degraded_guarantee_input / GHI_adjustment]
        return df_perf, [1] * length + [np.nan]*(12-length)
    
    #failed all three checks
    df_perf[['kwh_meas_poa', 'kwh_meas_ghi']] = [actual_production / POA_adjustment, actual_production / GHI_adjustment]
    df_perf['USB_result'] = df_perf[['kwh_meas', 'kwh_meas_poa', 'kwh_meas_ghi']].max(axis=1).item() / degraded_guarantee_input
    
    return df_perf, [0] * 12

def Regions_performance_guarantee(df, df_month_2, df_Pvsyst, df_Pvsyst_2_month, guarantee_input, df_perf):

    #    weights = df_Pvsyst_2_month['KWh_adj_by_days'] / sum([[i] * 3 for i in df_Pvsyst_2_month['KWh_adj_by_days'].resample("Q").sum()], [])
    #    df_quarterly = pd.DataFrame({"Project_IPR_%": (df_month_2['Project_IPR_%'] * weights).resample("Q").sum()})
    #    df_quarterly["Project_OPR_%"] = (df_month_2['Project_OPR_%'] * weights).resample("Q").sum()
    #
    #    df_perf[['reg_IPR', 'reg_OPR']] = df_quarterly[["Project_IPR_%", "Project_OPR_%"]]
    #    df_perf['reg_target'] = .95
    #    
    #    result = [1,1,1,1]
    #    for quarter in range(len(df_quarterly)):
    #        if df_quarterly.iloc[quarter]['Project_IPR_%'] > 0.95:
    #            result[quarter] = 1
    #        elif df_quarterly.iloc[quarter]['Project_OPR_%'] > 0.95:
    #            result[quarter] = 1
    #        else:
    #            result[quarter] = 0
    #    
    #    df_perf['reg_result'] = result
    
    #df_perf, guarantee_result = Regions_performance_guarantee(df, df_month_2, df_Pvsyst, df_Pvsyst_2_month, guarantee_input, df_perf) #from SPA to see inputs on 11/27
    
    #do over
    year = df_month_2.index.year[0] 
    filter_date = datetime.date(year, guarantee_input.month, guarantee_input.day) + pd.offsets.QuarterEnd() #get to end of the calendar year quarter using the input guarantee date
    
    # calculate how much each month contributes to yearly POA
    monthly_weights = df_Pvsyst_2_month['Year 0 Actual Production (kWh)'] / df_Pvsyst_2_month['Year 0 Actual Production (kWh)'].sum()
    monthly_weights.name = 'monthly_weights'
    # calculate what percentage of each month we have
    monthly_fractions = df['POA_avg'].resample("M").count() / df_Pvsyst['POA (W/m2)'].resample("M").count()
    monthly_fractions = monthly_fractions.fillna(0)
    monthly_fractions.name = 'monthly_fractions'

    #concat dataframes to join monthly fractions and weights with the IPR and OPR % values
    aux = pd.concat([df_month_2[['Project_IPR_%', 'Project_OPR_%']], monthly_weights, monthly_fractions], axis = 1)
    
    #montly filtered dataframes
    beg = aux.loc[aux.index <= filter_date, :]
    end = aux.loc[aux.index >  filter_date, :]
    
    #collapse into two rows
    beg_2 = Regions_math(beg)
    end_2 = Regions_math(end)
    
    aux_2 = beg_2.append(end_2)
    
    #evaluate test
    result = [1,1]
    for quarter in range(2):
        if aux_2.iloc[quarter]['Project_IPR_%'] > 0.95:
            result[quarter] = 1
        elif aux_2.iloc[quarter]['Project_OPR_%'] > 0.95:
            result[quarter] = 1
        else:
            result[quarter] = 0

    
    
    aux['result_monthly'] = -1
    aux.loc[beg.index, 'result_monthly'] = result[0]
    aux.loc[end.index, 'result_monthly'] = result[1]
    
    disp_date = filter_date + pd.offsets.Day(1)
    df_perf.loc['1/1/{} to {}'.format(year, disp_date.strftime('%m/%d/%Y')), ['reg_IPR', 'reg_OPR']] = beg_2[['Project_IPR_%', 'Project_OPR_%']].values
    df_perf.loc['{} to 12/31/{}'.format(disp_date.strftime('%m/%d/%Y'), year), ['reg_IPR', 'reg_OPR']] = end_2[['Project_IPR_%', 'Project_OPR_%']].values
    #df_perf.loc['12/31/{} to {}'.format(year, filter_date.strftime('%m/%d/%Y')), ['reg_IPR', 'reg_OPR']] = end_2[['Project_IPR_%', 'Project_OPR_%']].values

    df_perf['reg_result'] = result
    df_perf['reg_target'] = .95
    df_perf['Guar_freq'] = 'Annual'
    df_perf['Gaur_range'] = '{a}/{b} - {a}/{b}'.format(a=disp_date.month, b=disp_date.day)
    return df_perf, aux['result_monthly'].tolist()[:len(df_month_2)]

def Regions_math(beg):
    beg_2 = pd.DataFrame([], index = [0], columns = beg.columns)
    
    #sumproduct IPR
    beg['dot_IPR'] = (beg['Project_IPR_%'] * beg['monthly_weights']) / beg['monthly_weights'].sum()
    beg_2.loc[0, 'Project_IPR_%'] = beg['dot_IPR'].sum()
    
    #sumproduct OPR
    beg['dot_OPR'] = (beg['Project_OPR_%'] * beg['monthly_weights']) / beg['monthly_weights'].sum()
    beg_2.loc[0, 'Project_OPR_%'] = beg['dot_OPR'].sum()
    
    beg_2 = beg_2[['Project_IPR_%', 'Project_OPR_%']]
    return beg_2
    
    
def Sol_performance_guarantee(df_month_2, df_Pvsyst_2_month, df_perf):
    weights = df_Pvsyst_2_month['KWh_adj_by_days'] / sum([[i] * 3 for i in df_Pvsyst_2_month['KWh_adj_by_days'].resample("Q").sum()], [])
    df_quarterly = pd.DataFrame({"Project_IPR_%": (df_month_2['Project_IPR_%'] * weights).resample("Q").sum()})
    df_quarterly["NREL_Weather_Adj_%"] = (df_month_2['NREL_Weather_Adj_%'] * weights).resample("Q").sum()
    df_quarterly["Sol_%"] = df_quarterly["Project_IPR_%"] / df_quarterly["NREL_Weather_Adj_%"] 
    
    df_perf['sol_IPR_weather'] = df_quarterly["Sol_%"]
    df_perf['sol_target'] = .95
    df_perf['Guar_freq'] = 'Quarterly'
    
    out = [0,0,0,0]
    s = list(df_month_2.resample('QS').sum().index)
    for i in range(4): 
        out[i] = '{} - {}'.format(s[i].strftime('%m/%d'), df_quarterly.index[i].strftime('%m/%d'))
    df_perf['Gaur_range'] = out

    
    result = [1,1,1,1]
    for quarter in range(len(df_quarterly)):
        if df_quarterly.iloc[quarter]['Sol_%'] > 0.95:
            result[quarter] = 1
        else:
            result[quarter] = 0
    
    df_perf['sol_result'] = result
    
    return df_perf, sum([[i] * 3 for i in result], [])[:len(df_month_2)]

def IDP_performance_guarantee(ava, df_perf):
    H = ava['AVA'].resample('m').count()
    N = len([i for i in ava.columns if 'Inv_kw_' in i])
    DH = (N - ava.loc[ava['POA_avg'] > 50, 'Inv_ON']).resample('m').sum()
    #DH['2021-07-31'] = 529
    MA = (H * N - DH) / (H * N)
    
    # see section 6.4, do we need to do 75% for the first year?
    result = pd.Series(data = 0, index = pd.date_range(str(ava.index[0].year), periods=12, freq='m'))
    result[MA.index] = 1 * (MA.values > 0.85)
    
    month_start = result.resample("MS").mean().index.tolist()    

    df_perf['IDP_pass'] = result    
    df_perf['IDP_mechanical_ava'] = MA
    df_perf['IDP_target_ava'] = 0.85
    df_perf['Guar_freq'] = 'Monthly'
    df_perf['Gaur_range'] = ['{} - {}'.format(s.strftime('%m/%d'), e.strftime('%m/%d')) for s,e in zip(month_start, result.index.tolist())]
    
    return df_perf, result.tolist()


#added by SPA on 11/28/18
def SCEG_performance_guarantee(project_name, df, df_Pvsyst, PIS_date, df_perf):    
    
    #read in Net Energy and REC Delivery Requirements - One large Excel file for all the SCEG sites taking different tabs
    #df_attachement = pd.read_excel(r'D:\Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project\Python_Functions\Other Scripts\Testing - SCEG Performance Guarantees\SCEG_Perf Gaur_Contract_Quantity.xlsx', sheet_name=project_name) #NEED TO PUT THIS EXCEL FILE IN CENTRAL LOCATION AND RENAME
    df_attachement = pd.read_excel(os.path.join(file_project, 'Python_Functions\Other Scripts\Performance Guarantee Inputs\SCEG\SCEG_Perf Gaur_Contract_Quantity.xlsx'), sheet_name=project_name) #NEED TO PUT THIS EXCEL FILE IN CENTRAL LOCATION AND RENAME

    #get start year to test if we are in year 2 or not
    perf_guar_start_year = PIS_date.year
    
    #determine reporting year 
    reporting_year = df.index.year[-1]    

    #test to see if in year 1 of contract. logic is if the start year based on PIS date is not equal to current year. this will mess up when no PIS date is entered.
    #   if in year 2 or more...do the perf guarantee (contract states that perf guar occurs starting in year 2)
    if perf_guar_start_year!=reporting_year:

        ########################################################################################################
        #get proration - from RB
        # calculate how much each month contributes to yearly POA
        monthly_weights = df_Pvsyst['POA (W/m2)'].resample('M').sum() / df_Pvsyst['POA (W/m2)'].sum()
        
        # calculate what percentage of each month we have
        monthly_fractions = df['POA_avg'].resample("M").count() / df_Pvsyst['POA (W/m2)'].resample("M").count()
        monthly_fractions = monthly_fractions.fillna(0)
        proration = (monthly_weights * monthly_fractions).sum()
        ########################################################################################################

        #get contract_year and performance guarentee value
        contract_year = (reporting_year - perf_guar_start_year) + 1 #note the +1 as year 0 is actually contract year 1, so adding 1.
        net_energy = df_attachement.loc[df_attachement['Contract Year']==contract_year, 'Net Energy (kWh)'].item()

        #find out lost kWh due to grid outages TO subtract out of the Net Energy value which we compare on
        #   contract states 'provided that the Contract Quantity for Net Energy shall be reduced on an equitable basisa 
        grid_loss = df['Meter_&_ava_&_grid'].sum() - df['Meter_&_ava'].sum()
        
        #subtract grid loss from net_energy given contractual language, then with that value, derive the 85% target value
        net_energy_adj = net_energy - grid_loss
        net_energy_target = net_energy_adj * 0.85 * proration
        
        #compare our kWh-sum-to-date with the adjusted target
        if df['Meter_Corrected_2'].sum() >= net_energy_target:
            result = 1
        else:
            result = 0
            
        guarantee_result = [result] * 12

        #set up df_perf
        df_perf.loc[df.index.date[-1], ['SCEG_kwh_guar', 'SCEG_kwh_guar_grid_85_proration', 'SCEG_kwh_measured', 'SCEG_kwh_proration_factor', 'SCEG_perf_guar_result']] = [net_energy, net_energy_target, df['Meter_Corrected_2'].sum(), proration, result]

    else:
        guarantee_result = [1] * 12 #adjusted from ['null'] * 12 to adjust for datatype of SQL column

                    
    return df_perf, guarantee_result


# -*- coding: utf-8 -*-
"""
Created on Wed Aug 16 09:38:55 2017

@author: Kevin Anderson
"""

import numpy as np
import pandas as pd
import os
from sqlalchemy import create_engine
from win32com.client import Dispatch

username = os.path.split(os.path.expanduser("~"))[1]
if username == 'Kevin Anderson':
    # back in my day, C drives were good enough for anyone
    file_project = os.path.join(r'C:\Users', username, r'Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project')
elif username == 'blumenthal':
    file_project = r'C:\Users\blumenthal\Cypress Creek Renewables\AM-Performance - _Dashboard_Project'
    
elif username == 'Ryan':
    # the co-worker so nice, we elif'd him twice
    file_project = r'E:\Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project'

elif username == 'MartinWaters':
    file_project = r'C:\Users\MartinWaters\Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project'
    
elif username == 'EricFitch':
    # RIP
    file_project = r'C:\Users\EricFitch\Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project'
    
elif username == 'ChristopherDowns':
    #file_project = r'D:\Cypress Creek Renewables\AM-Performance - Documents\_Dashboard_Project'
    file_project=r'C:\Users\ChristopherDowns\Cypress Creek Renewables\AM-Performance - Documents\_Dashboard_Project'
    
elif username == 'PerfEng':
    file_project = r'C:\Users\PerfEng\Cypress Creek Renewables\AM-Performance - Documents\_Dashboard_Project'
    
elif username == 'corey.pullium':
    file_project = r'C:\Users\corey.pullium\Cypress Creek Renewables\AM-Performance - Documents\_Dashboard_Project'
    
elif username == 'MelissaFrench':
    file_project = r'C:\Users\MelissaFrench\Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project'

elif username == 'StoneHayden':
    file_project = r'C:\Users\StoneHayden\Cypress Creek Renewables\AM-Performance - _Dashboard_Project'
    
elif username == 'PradeepAmireddy':
    file_project = r'C:\Users\PradeepAmireddy\Cypress Creek Renewables\AM-Performance - _Dashboard_Project'
    
elif username == 'AnnaSchmackers':
    file_project = r'C:\Users\AnnaSchmackers\Cypress Creek Renewables\AM-Performance - Documents\_Dashboard_Project'

elif username == 'LukeSain':
    file_project = r'C:\Users\LukeSain\OneDrive - Cypress Creek Renewables\Documents - AM-Performance\_Dashboard_Project'

elif username == 'AndrewCurthoys':
    file_project = r'C:\Users\AndrewCurthoys\Cypress Creek Renewables\AM-Performance - Documents\_Dashboard_Project'


else:
    # saurabh because he likes the D drive better. 
    # Wait...why do I have to be the 'else'. I demand to be the 'if'
    # dude you put yourself there, idk what you're complaining about
    # malarky! there is a plot against me. and I don't mean one of the dashboard variety.
    # typical asset-management performance engineer, playing the victim when he tilted the POA himself
    # RIP
    file_project = r'D:\Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project'

file_production_data = os.path.split(file_project)[0]
    
def get_sql_engine():
    
    password_file = r'C:\Users\{}\Documents\Bartertown.txt'.format(username)
    
    with open(password_file, 'r') as f:
        lines = f.readlines()
    
    sql_username, sql_password = [line.split("'")[1] for line in lines]
    
    host = 'bartertown.cbnrsntwaejm.us-west-2.rds.amazonaws.com'
    port = '5432'
    db = 'thunderdome'
    
    url = 'postgresql+psycopg2://{}:{}@{}:{}/{}'
    url = url.format(sql_username, sql_password, host, port, db)
    
    engine = create_engine(url)
    return engine

def get_sql_engine_cgd(username=username):
    
    password_file = r'C:\Users\{}\Documents\Bartertown.txt'.format(username)
    
    with open(password_file, 'r') as f:
        lines = f.readlines()
    
    sql_username, sql_password = [line.split("'")[1] for line in lines]
    
    host = 'bartertown.cbnrsntwaejm.us-west-2.rds.amazonaws.com'
    port = '5432'
    db = 'thunderdome'
    
    url = 'postgresql+psycopg2://{}:{}@{}:{}/{}'
    url = url.format(sql_username, sql_password, host, port, db)
    
    engine = create_engine(url,pool_pre_ping=True)
    return engine

# yo who commented this out?  ridic
# six months later:  I still think this should be in the main code
    
#def face():
#    from scipy.misc import face
#    import matplotlib.pyplot as plt
#    plt.figure(); plt.imshow(face())

def save_df_existing_Excel_ALL (file_name_save, df_list_save, sheet_name_save):
    #
    xl = Dispatch('Excel.Application')
    wb = xl.Workbooks.Open(file_name_save)
    #
    for i in range(0, len(df_list_save)):  
        df_PV = df_list_save[i]
        df_PV = df_PV.reset_index()
        ws = wb.Worksheets(sheet_name_save[i])
        StartRow = 2
        StartCol = 1
        ws.Range(ws.Cells(StartRow,StartCol),# Cell to start the "paste"
                 ws.Cells(StartRow+len(df_PV.index)-1,
                          StartCol+len(df_PV.columns)-1)).Value = df_PV.values      
        wb.RefreshAll()     
    # Saves the Workbook
    wb.Save()
    wb.Close(True) 
    # Closes Excel
    xl.Quit()

def get_df_keys():
    df = pd.read_excel(os.path.join(file_project,'Python_Functions','Data Download Scripts','Fleet_database.xlsx'), sheetname='Data')
    return df

def get_ss_as_df(sheet_id):
    # helper function to query the data in a SS and put it into a dataframme

    # pip install smartsheet-python-sdk
    import smartsheet
    # API_TOKEN is user-specific, so SS will block you from accessing 
    # sheets that you don't have permissions for
    API_TOKEN = os.environ['SMARTSHEET_TOKEN']

    # Initialize API client
    ss_client = smartsheet.Smartsheet(API_TOKEN)
    # by default, the SS functions will return error objects if you do something wrong 
    # (eg query a nonexistent sheet), instead of raising exceptions
    # if you ask me, that's crazy.  so we tell it to raise exceptions instead
    ss_client.errors_as_exceptions(True)

    # page_size (rowcount) defaults to 100, which is annoying
    # we want everything, so we have to get the number of rows in the sheet first
    # set it to zero initially so we don't get a bunch of unnecessary data
    sheet = ss_client.Sheets.get_sheet(sheet_id, page_size=0)
    row_count = sheet.total_row_count
    # now query the sheet again, using the actual row count
    sheet = ss_client.Sheets.get_sheet(sheet_id, page_size=row_count)
    
    # extract out the cell values into a df
    sheet_dict = sheet.to_dict()
    # nested list comprehensions?  woahhhh
    values = [[cell.get('value', None) for cell in row['cells']] for row in sheet_dict['rows']]
    # convert None to nan.  for some reason this was difficult to do with the final df
    values = [[val if val is not None else np.nan for val in row] for row in values]
    df = pd.DataFrame(values)
    
    df = df.dropna(how='all')
    
    # df has the data, but the column names are just numbers
    df.columns = [col.title for col in sheet.get_columns(include_all = True).data]
    # bring the index over
    #df.index = [int(i) for i in df['Column1']]  #often breaks
    df.index = range(len(df))
    del df['Column1']
    
    return df

#DUPLICATE CREATED BY SAURABH TO ADJUST FOR SSs THAT DO NOT HAVE A 'Column1' in them (e.g. EPC Allocation tracker')
def get_ss_as_df_spa(sheet_id):
    # helper function to query the data in a SS and put it into a dataframme

    # pip install smartsheet-python-sdk
    import smartsheet
    # API_TOKEN is user-specific, so SS will block you from accessing 
    # sheets that you don't have permissions for
    API_TOKEN = os.environ['SMARTSHEET_TOKEN']

    # Initialize API client
    ss_client = smartsheet.Smartsheet(API_TOKEN)
    # by default, the SS functions will return error objects if you do something wrong 
    # (eg query a nonexistent sheet), instead of raising exceptions
    # if you ask me, that's crazy.  so we tell it to raise exceptions instead
    ss_client.errors_as_exceptions(True)

    # page_size (rowcount) defaults to 100, which is annoying
    # we want everything, so we have to get the number of rows in the sheet first
    # set it to zero initially so we don't get a bunch of unnecessary data
    sheet = ss_client.Sheets.get_sheet(sheet_id, page_size=0)
    row_count = sheet.total_row_count
    # now query the sheet again, using the actual row count
    sheet = ss_client.Sheets.get_sheet(sheet_id, page_size=row_count)
    
    # extract out the cell values into a df
    sheet_dict = sheet.to_dict()
    # nested list comprehensions?  woahhhh
    values = [[cell.get('value', None) for cell in row['cells']] for row in sheet_dict['rows']]
    # convert None to nan.  for some reason this was difficult to do with the final df
    values = [[val if val is not None else np.nan for val in row] for row in values]
    df = pd.DataFrame(values)
    
    df = df.dropna(how='all')
    
    # df has the data, but the column names are just numbers
    df.columns = [col.title for col in sheet.get_columns(include_all = True).data]
    # bring the index over
    #df.index = [int(i) for i in df['Column1']]
    #del df['Column1']
    
    return df


    
# df_keys SS ID
sheet_id = '8659171076794244'
all_df_keys = get_ss_as_df(sheet_id)
all_df_keys.index = all_df_keys['Project']
for col in ['PIS', 'FC']:
    all_df_keys[col] = pd.to_datetime(all_df_keys[col])
    
df_keys = all_df_keys.loc[all_df_keys['Retired'] != True, :]

#df_keys = get_df_keys()

try:
    sql_engine = get_sql_engine()
except Exception as e:
    print('SQL Engine failed to load')




# -*- coding: utf-8 -*-

from logging_conf import get_logging_config
from datetime import datetime
import logging
import logging.config
from matplotlib.pyplot import waitforbuttonpress
from numbers import Number
import numpy as np
import os
import pandas as pd
from pandas import DataFrame
from pandas.io.sql import SQLTable
import scipy.stats as sct
import shutil
import sys
from typing import Union

from ccrenew.dashboard.data_processing.CCR import (
    all_df_keys,
    file_project,
)
from ccrenew.dashboard.plotting.plots import Plotter
from ccrenew.dashboard.project import Project
import ccrenew.dashboard.utils.dashboard_utils as dbutils
import ccrenew.dashboard.utils.project_neighbors as neighbs
from ccrenew.dashboard.utils.dashboard_utils import func_timer
from ccrenew.dashboard.utils.df_tools import df_update_join

    
# Python 2 compatibility
if sys.version_info.major == 3:
    unicode = str

# suppress warnings about "A value is trying to be set on a copy of a slice from a DataFrame."
pd.options.mode.chained_assignment = None

# Initialize logger
# Get logging configuration from the logging.conf file in the same directory as this file
username = os.getenv('username')
dashboard_folder = file_project
log_filename = '{}_dashboard-{}.log'.format(datetime.now().strftime('%Y-%m-%d_%H-%M-%S'), username)
log_config = os.path.join(dashboard_folder, 'Python_Functions', 'logging.conf')
log_file = os.path.join(dashboard_folder, 'Python_Functions', '_old', 'logs', log_filename)
LOGGING_CONFIG = get_logging_config(log_file)
logging.config.dictConfig(LOGGING_CONFIG)

# Create logger
logger = logging.getLogger(__name__)
logger.info('Beginning Dashboard Session')

def _execute_insert(self, conn, keys, data_iter):
    #print "Using monkey-patched _execute_insert"
    data = [dict((k, v) for k, v in zip(keys, row)) for row in data_iter]
    conn.execute(self.insert_statement().values(data))


SQLTable._execute_insert = _execute_insert

    
class DashboardSession(object):
    """Main orchestrator for processing, plotting, and exporting data for projects (i.e. sites).

    Args:
        project_list (str or list, optional): List of project(s) to initialize. Defaults to None.
        data_cutoff_date (str or datetime, optional): The last date you want to analyze. Defaults to the current date the dashboard is being run.
        data_source (str, optional): The data source to run the dashboard on. Defaults to "AE_data_".

    """

    # Instantiate globals as class variables
    dashboard_dir = file_project
    """Directory for the dashboard project"""
    df_keys = all_df_keys
    """Metadata for all projects"""

    # Site lists
    tracker_sites = ['5840 Buffalo Road', 'Alexis', 'ATOOD II', 'Bar D', 'Barnhill Road Solar', 'Bay Branch',
                     'Bonanza', 'Bovine Solar', 'Bronson', 'Cascade', 'Chisum', 'Copperfield', 'Curie',
                     'Eddy II', 'Gaston II', 'Griffin', 'Grove Solar (NC)', 'Grove', 'Hardison Farm',
                     'Hopewell Friends', 'Hyline', 'IS - 46 Tracker', 'IS 67', 'Lampwick', 'Leggett Solar',
                     'Mars', 'Neff', 'Nimitz', 'Open Range', 'Palmetto Plains', 'Prince George 1',
                     'Prince George 2', 'Railroad', 'Shoe Creek', 'Siler', 'Simba', 'Springfield', 'Sterling',
                     'Thunderegg', 'Vale', 'Wagyu', 'Warren', 'Wendell', 'West Moore II', 'Yellow Jacket']
    """Projects with tracker-type racking"""
    battery_ac_funds = ['LGE']
    """Funds that have projects with AC batteries"""
    battery_ac_sites = df_keys.query("Fund in @battery_ac_funds").index
    """Projects with AC batteries"""
    battery_dc_sites = ['Salt Point', 'Dubois']
    """Projects with DC batteries"""

    def __init__(self,
                 project_list=None,
                 data_cutoff_date=None,
                 data_source='AE_data_',
                 **kwargs):
        self.project_list = project_list
        """List of `Project` objects that have been initialized & added to the `DashboardSession` instance"""
        if not data_cutoff_date:
            self.data_cutoff_date = datetime.today()
        else:
            self.data_cutoff_date = data_cutoff_date
        self.data_source = data_source
        
            
        # Snow dataframe from Smartsheets
        self.raw_snow_df = dbutils.get_snow_df(dashboard_folder, data_source)
        """Snow data pulled from Smartsheets"""

        # Initialize a dict to store projects - key = project name; value = project object
        self.project_list = {}

        # Add projects to the session if provided
        if project_list:
            if isinstance(project_list, (str, unicode, list)):
                self.add_projects(project_list)
            else:
                print('`project_list` must be a string or list. The projects provided were not added to the DashboardSession. Please try again in the initialized DashboardSession.')
        
        # Create list for projects that have errored out
        self.errored_projects = {}

        # List of plotter objects if you want to have multiple going in your session
        self.plotters = {}


    def __str__(self):
        return 'DashboardSession object with {} projects'.format(len(self.project_list))


    @func_timer
    def add_project(self, project_name, get_neighbors=True):
        """Adds & initializes a project to a `DashboardSession` instance.

        Args:
            project_name (str or Project): The name of the project (i.e. solar site) to add to the `DashboardSession` instance.
            get_neighbors (bool, optional): Flag to add & initialize neighbor projects. Defaults to True.
        """
        # Type validation
        if not isinstance(project_name, (str, unicode, list)):
            raise TypeError("`project_name` must be a string or Python list! Please reformat input and try again.")

        # Determine behavior based on what was passed to method
        if isinstance(project_name, list):
            self.add_projects(project_name)
        elif project_name in self.project_list:
            return
        else:
            try:
                project = self.__initialize_project(project_name, get_neighbors)
                self.project_list[project.project_name] = project
            except RuntimeError as e:
                print(e)
                logger.warn(e)
    

    def add_projects(self, project_list, get_neighbors=True):
        """Calls `add_project` for each project in a list.

        Args:
            project_list (list, str): list of project names to add & initialize to the `DashboardSession` instance.
            get_neighbors (bool, optional): Flag to add & initialize neighbor sites. Defaults to True.
        """
        if isinstance(project_list, (str, unicode)):
            self.add_project(project_list, get_neighbors)

        elif isinstance(project_list, list):
            for project_name in project_list:
                self.add_project(project_name, get_neighbors)
        
        else:
            raise TypeError("`project_list` must be a string or Python list! Please reformat input and try again.")
                

    @func_timer
    def process_project(self, project_name, reprocess=False):
        """Processes data for the given project.

        Args:
            project_name (str, Project, or list): The name of the project to process.
            reprocess (bool, optional): Whether to reprocess the data. This option will likely\
            be uncommon as any changes to a project's config file or powertrack file will\
            automatically reprocess that data. This could be used if any changes are made\
            outside of the config/powertrack file or a change is made to a neighbor. Defaults to False.
        """
        if isinstance(project_name, list):
            self.process_projects(project_name, reprocess)
        elif isinstance(project_name, (str, unicode, Project)):
            # Get project object
            project = self.__get_project(project_name)

            # Pull the last updated dates from the filesystem for the config & Powertrack files
            last_update_config_file = os.path.getmtime(project.config_filepath)
            last_update_powertrack_file = os.path.getmtime(project.powertrack_filepath)

            # Check if the config file or the Powertrack file has been updated.
            # If not we don't need to process.
            if (
                project.last_update_config != last_update_config_file or
                project.last_update_powertrack != last_update_powertrack_file
                ):
                # Pull config and/or Powertrack data
                self.__prepare_source_data(project)
            else:
                if project.processed and not reprocess:
                    print('{} already processed'.format(project.project_name))
                    return True

            # Process data - update neighbor sensors first then process the project
            self.__get_neighbor_sensor_data(project)
            project._process_data()
            project.processed = True
            print('{} successfully processed'.format(project.project_name))
        else:
            raise TypeError("`project_name` must be a string or Project object! Please reformat input and try again.")
    
    
    def process_projects(self, project_list, reprocess=False):
        """Calls `process_project` for each project in a list

        Args:
            project_list (list, str, or Project): List of project names of `Project` objects to process.
            reprocess (bool, optional): See `process_project`. Defaults to False.
        """
        if isinstance(project_list, (str, unicode, Project)):
            self.process_project(project_name, reprocess)
        elif isinstance(project_list, list):
            for project_name in project_list:
                self.process_project(project_name, reprocess)
        else:
            raise TypeError("`project_list` must be a string or Python list! Please reformat input and try again.")
    

    def process_all_projects(self):
        """Processes all projects in the `project_list` for the `DashboardSession` instance
        """
        # FIXME: add data cutoff date in here in case you want to change that in the middle of a session
        # FIXME: Same with process project object below
        project_list = self.project_list.keys()
        self.process_projects(project_list)

    
    def __get_project(self, project_name, get_neighbors=True):
        """Returns a project object given a project name.

        Args:
            project_name (str or Project): Project name to add to the `DashboardSession` instance.
            get_neighbors (bool, optional): Flag to add & initialize neighbor projects. Defaults to True.

        Returns:
            Project: the initialized `Project` given the project name.
        """
        if isinstance(project_name, Project):
            return project_name
        try:
            project = self.project_list[project_name]
        except KeyError:
            self.add_project(project_name, get_neighbors)
            project = self.project_list[project_name]
        
        return project


    def __prepare_source_data(self, project):

        # Update config file
        # If the config file hasn't been updated it will use the data that's already been pulled
        # If the config file has been updated it will read the file & update the data
        self.__update_project_config(project)

        # Load powertrack data
        # If the powertrack file hasn't been updated it will use the data that's already been pulled
        # If the powertrack file has been updated it will read the file & update the data
        project.load_production_data()
    

    def __update_project_config(self, project):
        
        # Store the list of sensors before pulling the config file
        sensors = project.Get_Sensor
        project._parse_config_file()

        # Check if new neighbor sensors have been added to the config & pull them if so
        if not sensors.equals(project.Get_Sensor):
            project._find_neighbor_sensors(self)
            self.__get_neighbor_sensor_data(project)
            
            
    def __get_neighbor_sensor_data(self, project):
        for neighbor_name in project.neighbor_sensors:
            # Update production data from neighbor
            try:
                neighbor = self.__get_project(neighbor_name, get_neighbors=False)
            except KeyError:
                print('Neighbor for {} not found: {}. Sensor data for the neighbor will be blank. Check logs to determine why neighbor was not added to DashboardSession'.format(project.project_name, neighbor_name))
            
            # Reload neighbor's config or powertrack files if needed
            self.__prepare_source_data(neighbor)

            # Find the columns needed from the neighbor
            sensor_cols = project.Get_Sensor.loc[project.Get_Sensor['Source'] == neighbor_name, 'Value'].tolist()

            # Get the columns from the neighbor
            try:
                neighbor_df = neighbor.df.reindex(index=project.df.index, columns=sensor_cols)
            # If the production data hasn't been loaded for the project
            except AttributeError:
                neighbor.load_production_data()
                neighbor_df = neighbor.df.reindex(index=project.df.index, columns=sensor_cols)
            # If the neighbor errored out during initialization we'll just create a blank df for the neighbor
            except NameError:
                neighbor_cols = [col + '_' + neighbor_name for col in sensor_cols]
                project.df = project.df.reindex(columns = project.df.columns.tolist() + neighbor_cols)
                project.df_sensor_ON = project.df_sensor_ON.reindex(columns = project.df_sensor_ON.columns.tolist() + neighbor_cols)
                return

            neighbor_df.rename(columns=lambda x: np.str(x) + '_' + neighbor_name, inplace=True)
            neighbor_sensor_ON = neighbor.df_sensor_ON.reindex(index=project.df_sensor_ON.index, columns=sensor_cols)
            neighbor_sensor_ON.columns = neighbor_df.columns.tolist()

            df_cols = project.df.columns.tolist()
            project.df = df_update_join(project.df, neighbor_df)
            project.df = project.df[df_cols + neighbor_df.columns.tolist()]

            project.df_sensor_ON = df_update_join(project.df_sensor_ON, neighbor_sensor_ON)
            project.df_sensor_ON = project.df_sensor_ON[project.df.columns.tolist()]

    
    def draw_plots(self, project_name, plot_order=None, *args, **kwargs):
        """Draws plots for the given project

        Args:
            project_name (str, Project, or list): A project to draw plots on for analysis.
            plot_order (list or str, optional): A list of plots to draw. This can be
                any number of plots from one to all plots. Defaults to all plots in the below order.
            <h2>Acceptable Options & Default Order:</h2>
            * `xplot_met_poa`: Crossplot of meter & POA data.
            * `xplot_temp`: Crossplot of meter & POA, colored by Tamb.
            * `temps`: Tcell temperature comparison.
            * `inv`: Inverters.
            * `pr`: Hourly PR plot.
            * `8760`: Measured vs 8760 comparison.
            * `weather`: Weather sensors.
            * `met_corr`: Meter corrections.
            * `met_corr_dif`: Meter correction dif.
            * `poas`: POA sensors.
            * `met_poa`: Meter & POA avg.
            * `ghi`: GHI sensors.
            * `irrad`: POA & GHI sensors.
            * `tz`: Timezone check.
            * `losses`: Losses by type.
            * `poa_corr`: POA correlation check.
        """
        if isinstance(project_name, (str, unicode, Project)):
            project = self.__get_project(project_name, get_neighbors=False)
            redraw = kwargs.get('redraw', False)
            already_processed = self.process_project(project_name)
            if already_processed and not redraw:
                # Check if the project has an active plotter object & if the plot order is the same
                if project.plotter and project.plotter.plot_list == plot_order:
                    print('Plots already drawn. Add a new plot to the `plot_list` attribute or set `redraw` = True if you\'d like to redraw the current plots')
                    return
                else:
                    self.__draw_plots(project, plot_order, *args, **kwargs)
            else:
                self.__draw_plots(project, plot_order, *args, **kwargs)
        elif isinstance(project_name, list):
            project_list = project_name
            for i, project_name in enumerate(project_list):
                project = self.__get_project(project_name, get_neighbors=False)
                self.draw_plots(project.project_name, *args, **kwargs)
                try:
                    print('Press any key in the last plot window to show plots for {}'.format(project_list[i+1]))
                except IndexError:
                    print('Last plot in the list. Press any key in the last plot window to exit loop. Plots will remain showing until you close them.')
                while True:
                    if waitforbuttonpress():
                        break

    
    def __draw_plots(self, project, plot_order, *args, **kwargs):
        """Private method to orchestrate drawing of plots

        Args:
            project (Project): A project to use to draw plots for analysis
        """
        project_name = project.project_name
        close_plots = kwargs.get('close_plots', True)
        if not project.plotter:
            project.plotter = Plotter(project, *args, **kwargs)
        if close_plots:
            project.plotter.close_plots()
        plotter_init_time = datetime.now().strftime('%m/%d/%Y %H:%M:%S')
        self.plotters = {project_name: project.plotter}
                
        project.plotter.draw_dashboard_plots(plot_order, *args, **kwargs)

    
    def __initialize_project(self, project_name, get_neighbors=True):
        """Initializes a project object with metadata

        Args:
            project_name (str): Name of a project to initialize.
            get_neighbors (bool, optional): Flag to add & initialize neighbor projects. Defaults to True.

        Returns:
            Project: An initialized Project object
        """

        try:
            df_proj_keys = self.df_keys.query("Project == @project_name").to_dict('records')[0]
        except IndexError:
            # Log that project was not found
            warn_msg = '"{}" not found in df keys. Please verify the project is present in df keys & the spelling is correct and try again'.format(project_name)
            print(warn_msg)
            logger.warn(warn_msg)

            # Remove project from project_list since we won't be able to initialize it
            del self.project_list[project_name]
            return

        if get_neighbors:
            # Find all neighbors that satisfy the distance and equipment requirements
            neighbor_list = neighbs.find_nearby_similar_projects(project_name, print_data=False).index.tolist()
            
            # Then remove the search project from the list
            # (Sometimes it will return an empty list if there is some data missing in `df_keys`) so we need to check if it returned a non-empty list first
            try:
                neighbor_list.remove(project_name)
            except ValueError:
                pass

            # If the project has neighbors, we'll add them to the DashboardSession instance
            # Set get_neighbors to False so we don't get neighbors of neighbors of neighbors etc.
            if neighbor_list:
                self.add_projects(neighbor_list, get_neighbors=False)
        else:
            neighbor_list = None

        Battery_AC_site = project_name in self.battery_ac_sites
        Battery_DC_site = project_name in self.battery_dc_sites
        Tracker_site = project_name in self.tracker_sites
        
        proj_init_dict = {}
        proj_init_dict['project_name'] = project_name
        proj_init_dict['df_proj_keys'] = df_proj_keys
        proj_init_dict['dashboard_dir'] = self.dashboard_dir
        proj_init_dict['data_cutoff_date'] = self.data_cutoff_date
        proj_init_dict['data_source'] = self.data_source
        proj_init_dict['Battery_AC_site'] = Battery_AC_site
        proj_init_dict['Battery_DC_site'] = Battery_DC_site
        proj_init_dict['Tracker_site'] = Tracker_site
        proj_init_dict['raw_snow_df'] = self.raw_snow_df
        proj_init_dict['neighbor_list'] = neighbor_list

        project = Project(proj_init_dict)
        return project
        

DashboardSession.__annotations__ = {'battery_ac_funds':list,
                                    'battery_ac_sites':list,
                                    'battery_dc_sites':list,
                                    'tracker_sites': list,
                                    'dashboard_dir': Union[str, bytes],
                                    'df_keys': DataFrame,
                                    'raw_snow_df': DataFrame,
                                    'project_list': list}



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 24 15:18:49 2021

@author: Chris Downs
"""
###Default Imports##################################################################################
import os
import numpy as np
import pandas as pd
import pvlib
from pvlib.location import Location
from multiprocessing.pool import ThreadPool

from ccrenew.dashboard.data_processing import solcats_API as Cats 
from ccrenew.dashboard.data_processing.CCR import (
    file_project,
    all_df_keys as df_keys
)
'''
purpose:
    support a library of functions that will allow for the transposition of ghi data (or data from a weather api) to POA data.
    include functions to plot calibration metrics (how well does transposed match actual?) how well does weather api match actual ghi, etc
'''
#%%                    ##Thoughts
'''
heiarchy of irradiance data
    1: native functioning POA
    2: neighbor functioning poa from a matching site (within 10 miles, same tilf or gcr, same tracking tech (back vs true))
    3: native site ghi data transposed ot POA (5 MINUTE DATA TRANSPOSED AND RESAMPLED TO HOURLY) dashboard will have an hourly ghi data plot, but transpose on more granular data
    4: neighbor GHI site (doesnt have to match) 5 minute data transposed to poa nd hourly resample
    5: weather API data transposed and resampled to hourly data

'''

#%%                    ##Functions
def simulate_trackers(project_name, st, ed, res="1Min"):
   
    #get infor from df_keys    
    lat, lon = df_keys.loc[project_name, ['GPS_Lat', 'GPS_Long']]
    tz = df_keys.loc[project_name, 'Timezone']
    gcr = df_keys.loc[project_name, 'Tilt/GCR']
    altitude=df_keys.loc[project_name,'Elevation']
    max_angle=df_keys.loc[project_name,'Max_angle']    
    
    temp_coef=df_keys.loc[project_name,'Temp_Coeff_Pmax']
    
    #choose tracking algorithm based upon temp_coefficient of the modules
    backtrack=True
    if temp_coef > -0.39:
        backtrack=False

    tz = str("US/" + tz) # can't be a unicode string >:(
    
    # altitude in meters above sea level.  name is optional, but helpful for documentation
    loc = Location(lat,	lon, tz=tz, altitude=altitude, name=project_name)
    
    # may want to change the freq, 1min is kind of heavy
    times = pd.date_range(start=st, end=ed, freq=res, tz = loc.tz)
    
    solpos = pvlib.solarposition.get_solarposition(times, loc.latitude, loc.longitude)
    
    tracker_data = pvlib.tracking.singleaxis(solpos['apparent_zenith'], solpos['azimuth'],
                                         axis_tilt=0, axis_azimuth=180, max_angle=max_angle,
                                         backtrack=backtrack, gcr=gcr).tz_localize(None)
    
    #I don't want Na values filled to 0. For ava purposes, 
    #tracking matters when model is not nan, so filters are applied later using this fact
    #tracker_data=tracker_data.fillna(0)
    
    return tracker_data

def half_interval_up(times):
    """
    shift a Datetimeindex forward by half an interval
    """
    shift_quantity=(times[1]-times[0])/2. #get what half a shift would be
    times+=shift_quantity
    return times

def half_interval_down(times):
    """
    shift a Datetimeindex forward by half an interval
    """
    shift_quantity=(times[1]-times[0])/2. #get what half a shift would be
    times-=shift_quantity
    return times

def transpose(surface_tilt, surface_azimuth, albedo, solar_position,
              ghi, dni, dhi, dni_et, airmass, **kwargs):
    """
    Estimate POA irradiance. PVWatts v5 uses a similar but slightly different
    method for near-horizon diffuse irradiance.

    Parameters
    ----------
    surface_tilt, surface_azimuth : numeric
        Array orientation [degrees]
    albedo : numeric
        Ground albedo
    solar_position : pd.DataFrame
        Solar position
    ghi, dni, dhi : numeric
        The three irradiance components
    airmass : numeric
        relative airmass
    dni_et : numeric
        extraterrestrial (top of atmosphere) irradiance
    **kwargs
        Extra arguments passed to ``pvlib.irradiance.get_total_irradiance``.

    Returns
    -------
    poa : pd.DataFrame
        POA irradiance components, including ``poa_global``.
    """
    poa = pvlib.irradiance.get_total_irradiance(
        surface_tilt,
        surface_azimuth,
        solar_position['apparent_zenith'],
        solar_position['azimuth'],
        dni,
        ghi,
        dhi,
        albedo=albedo,
        airmass=airmass,
        dni_extra=dni_et,
        model='perez',
        **kwargs)
    # return the entire dataframe, not just poa_global, for component calculations
    return poa.fillna(0)

def decompose(df,lat,lon,elevation,tz):
    """
    decompose the given ghi dataframe into both dhi and dni
    
    Parameters
    ----------
    df: pandas df of ghi data
    lat: Numeric. site latitude as a decimal
    lon: Numeric. site longitude as a decimal
    
    Returns
    -------
    dni
    dhi
    airmass
    dni_et
    """
    #df is a dataframe of ghi data
    
    
    
    #site_location = pvlib.location.Location(lat, lon,altitude=elevation,tz=tz)
    
    site_location = Location(lat,	lon, tz=tz, altitude=elevation)
    times = pd.date_range(start=df.index[0], end=df.index[-1], freq='5min', tz = site_location.tz)
    solar_position = site_location.get_solarposition(times=df.index)
    dni_et = pvlib.irradiance.get_extra_radiation(df.index)
    airmass = site_location.get_airmass(df.index)['airmass_relative']

    zenith=solar_position.apparent_zenith
    #times=df.index  #im pretty sure this line isnt used
    erb=pvlib.irradiance.erbs(df, zenith, times)#GHI data, zenith, datetimes
    dni = erb['dni']
    dhi = erb['dhi']
    ghi=df
    return ghi,dni,dhi,airmass,dni_et,site_location,solar_position

def get_solcast_data(ccr_id,start,end):
    """
    get data off s3 for solcast data
    
    Parameters
    ----------
    ccr_id: (str) ccr_id as a string
    start: (str) timestamp in form '2022-4-3' for the start of the data request
    end: (str) timestamp in form '2022-4-4' for the end of the data request
    
    Returns
    -------
    df_cat: (df) df of the weather from solcast to nearest day
    """
    print(type(start))
    if type(start)==str: start=start.split(' ')[0]
    else: start=start.date()
    if type(end)==str: end=end.split(' ')[0]
    else: end=end.date()
    
    date_range=pd.date_range(start,end)
    keys = [Cats.make_key(ccr_id, date) for date in date_range]
    
    pool_size = 6
    pool = ThreadPool(pool_size)
    day_df_list = pool.map(Cats.retrieve_df, keys)
    pool.close()
    pool.join()
    
    df_cat = pd.concat(day_df_list, axis=0)
    
    return df_cat

def solcats_to_dash(project,start,end,resample=True,convert=True):
    """
    parse out solcast data and turn it into poa useable for the dashboard
    
    Parameters
    ----------
    project: (str) string name  for the site from df_keys. same used in project list for the dashboard
    start: (str) timestamp in form '2022-4-3' for the start of the data request
    end: (str) timestamp in form '2022-4-4' for the end of the data request
    resample: (Bool) true to go to hourly, false to stay at 5 minute
    
    Returns
    -------
    df_poa: (df) df of the POA,Tamb, and windspeed values from solcast with units changed to the right format for project
    df_units: (df) just a list of units used since they are converted 
    """
    #get all that ish from df_keys
    ccr_id=df_keys.CCR_ID.loc[df_keys.Project==project].item()
    tz='US/'+str(df_keys['Timezone'].loc[df_keys['Project']==project].item())
    lat=df_keys.GPS_Lat.loc[df_keys['Project']==project].item()
    lon=df_keys.GPS_Long.loc[df_keys['Project']==project].item()
    racking=df_keys.Racking.loc[df_keys['Project']==project].item()
    elevation=df_keys['Elevation'].loc[df_keys['Project']==project].item()
    max_angle=df_keys['Max_angle'].loc[df_keys['Project']==project].item()
    temp_coef=df_keys['Temp_Coeff_Pmax'].loc[df_keys['Project']==project].item()
    axis_tilt=0
    albedo=1
    #pull in the solcast data:
    df_cat= get_solcast_data(ccr_id,start,end)
    
    #do some cheeky stuff that is racking specific
    if racking =='Fixed':
        tracking=False
        tilt_value=df_keys['Tilt/GCR'].loc[df_keys['Project']==project].item()
        surface_azimuth_value=180
        
        #make even fixed tilts a df of values so every time slot has a value and tz localize works the same for fixed vs trackers
        tilt_data=df_cat.copy()
        tilt_data['surface_tilt']=tilt_value
        tilt_data['surface_azimuth']=surface_azimuth_value
        surface_azimuth=tilt_data.surface_azimuth
        tilt=tilt_data.surface_tilt
        gcr=None
    elif racking =='Tracker':
        tracking=True
        gcr=df_keys['Tilt/GCR'].loc[df_keys['Project']==project].item()
        
        tracker_data=simulate_trackers(project, df_cat.index[0], df_cat.index[-1], res="5Min")
        surface_azimuth=tracker_data.surface_azimuth
        tilt=tracker_data.surface_tilt
  
    tilt=tilt.tz_localize(tz)
    df_cat=df_cat.tz_localize(tz)
    surface_azimuth=surface_azimuth.tz_localize(tz)
    #decompose so everything is consistent, but just use the regular dni and dhi in solcats
    ghi,dni,dhi,airmass,dni_et,site_location,solar_position=decompose(df_cat['ghi'],lat,lon,elevation,tz) 
    dni=df_cat['dni']
    dhi=df_cat['dhi']
    
    poa_df=transpose(tilt, surface_azimuth, albedo, solar_position, ghi, dni, dhi, dni_et, airmass) 
    df_cat['poa']=poa_df['poa_global']
    df_units=pd.DataFrame(data={'Temp':['Celsius'],'Wind':['m/s']})
    
    #add in a tmod column based on the conversion equation
    #Tmod = Tamb + POA *np.exp( a+ b*wind_speed)
    a_module=df_keys['a_module'].loc[df_keys['Project']==project].item()
    b_module=df_keys['b_module'].loc[df_keys['Project']==project].item()
    df_cat['Tmod']=df_cat['air_temp']+(df_cat['poa']*np.exp(a_module+b_module*df_cat['wind_speed_10m']))
    
    
    
    if convert:
        #do conversion of units
        folder=df_keys.Folder.loc[df_keys.Project==project].item()
        config_file =  project + r'_Plant_Config_MACRO.xlsm'
        config_path= os.path.join(file_project,folder,project,'Plant_Config_File',config_file)
        xl=pd.ExcelFile(config_path)
        df_config=xl.parse('Unit_Tab')
        
        #temp units
        temps=df_config.loc[df_config['Var_ID'].str.contains('amb')]['Convert_Farenheit_to_Celcius'].sum()
        temps+=df_config.loc[df_config['Var_ID'].str.contains('mod')]['Convert_Farenheit_to_Celcius'].sum()
        if temps >0:
            df_cat['Tamb']=df_cat['air_temp']*float((9/5.))+32.0
            df_units['Temp']='Fahrenheit'
            df_cat['Tmod']=df_cat['Tmod']*float((9/5.))+32.0
        else:
            df_cat['Tamb']=df_cat['air_temp']
            
        #wind units
        winds=df_config.loc[df_config['Var_ID'].str.contains('speed')]['Convert_mph_to_mps'].sum()
        if winds>0:
            df_cat['Wind_speed']=df_cat['wind_speed_10m']*float(2.23694)
            df_units['Wind']='Mph'
        else:
            df_cat['Wind_speed']=df_cat['wind_speed_10m']
    else:
        df_cat['Wind_speed']=df_cat['wind_speed_10m']
        df_cat['Tamb']=df_cat['air_temp']
    
    #do resample
    if resample:
        df_cat=df_cat.resample('H').mean()
        #correct for static values that solcats reports that the dashboard is gonna turn off.
        df_cat['Correction_factor']=np.where(df_cat.index.hour%2==0, 0.999, 1.001)
        df_cat['Tamb']*=df_cat['Correction_factor']
        df_cat['Tmod']*=df_cat['Correction_factor']
        df_cat['Wind_speed']*=df_cat['Correction_factor']
    return df_cat[['poa','Tamb','Tmod','Wind_speed','ghi']],df_units


def site_s3_to_poa(project,start,end,GHI_index=0,resample=True):
    """
    go get s3 data from s3 and tranpose to useable, pasteable POA data
    
    Parameters
    ----------
    project: (str) string name  for the site from df_keys. same used in project list for the dashboard
    start: (str) timestamp in form '2022-4-3' for the start of the data request
    end: (str) timestamp in form '2022-4-4' for the end of the data request
    GHI_index: (int) ghi iloc to use. 0 will use the first ghi in the list
    resample: (Bool) true to go to hourly, false to stay at 5 minute
    
    Returns
    -------
    df_poa: (df) df of the POA
    """
    #get all that ish from df_keys
    ccr_id=df_keys.CCR_ID.loc[df_keys.Project==project].item()
    tz='US/'+str(df_keys['Timezone'].loc[df_keys['Project']==project].item())
    lat=df_keys.GPS_Lat.loc[df_keys['Project']==project].item()
    lon=df_keys.GPS_Long.loc[df_keys['Project']==project].item()
    racking=df_keys.Racking.loc[df_keys['Project']==project].item()
    elevation=df_keys['Elevation'].loc[df_keys['Project']==project].item()
    max_angle=df_keys['Max_angle'].loc[df_keys['Project']==project].item()
    temp_coef=df_keys['Temp_Coeff_Pmax'].loc[df_keys['Project']==project].item()
    axis_tilt=0
    albedo=1
    
    
    #pull in the s3 data
    df=Cats.get_df(ccr_id,start,end,prod_data=True)
    #get just the one ghi column to use
    ghi_name=[col for col in df.columns if 'GHI' in col][GHI_index]
    
    df_cat=df[ghi_name]
    #do some cheeky stuff that is racking specific
    if racking =='Fixed':
        tracking=False
        tilt_value=df_keys['Tilt/GCR'].loc[df_keys['Project']==project].item()
        surface_azimuth_value=180
        
        #make even fixed tilts a df of values so every time slot has a value and tz localize works the same for fixed vs trackers
        tilt_data=df.copy()
        tilt_data['surface_tilt']=tilt_value
        tilt_data['surface_azimuth']=surface_azimuth_value
        surface_azimuth=tilt_data.surface_azimuth
        tilt=tilt_data.surface_tilt
        gcr=None
    elif racking =='Tracker':
        tracking=True
        gcr=df_keys['Tilt/GCR'].loc[df_keys['Project']==project].item()
        
        tracker_data=simulate_trackers(project, df_cat.index[0], df_cat.index[-1], res="5Min")
        surface_azimuth=tracker_data.surface_azimuth
        tilt=tracker_data.surface_tilt
  
    tilt=tilt.tz_localize(tz)
    df_cat=df_cat.tz_localize(tz)
    surface_azimuth=surface_azimuth.tz_localize(tz)
    
    #decompose so everything is consistent, but just use the regular dni and dhi in solcats
    ghi,dni,dhi,airmass,dni_et,site_location,solar_position=decompose(df_cat,lat,lon,elevation,tz) 
    poa_df=transpose(tilt, surface_azimuth, albedo, solar_position, ghi, dni, dhi, dni_et, airmass) 
    df_poa=poa_df['poa_global']
    
    #do resample
    if resample:
        df_poa=df_poa.resample('H').mean()
        
    return df_poa
    

"""
#%%                    ##Testing
plt.close('all')

project='Thunderegg'

df_data.index=df_data.Timestamp
df_data.index=df_data.index.to_datetime()

ghi_list=[col for col in df_data if 'GHI' in col and 'Dark' not in col]

poa_list=[col for col in df_data if 'POA' in col and 'Dark' not in col]


tz='US/'+str(df_keys['Timezone'].loc[df_keys['Project']==project].item())

df_data.index=df_data.index.tz_localize(tz)
df_data[ghi_list].plot()



#stored sitelevel data
lat=df_keys.GPS_Lat.loc[df_keys['Project']==project].item()
lon=df_keys.GPS_Long.loc[df_keys['Project']==project].item()
racking=df_keys.Racking.loc[df_keys['Project']==project].item()


if racking =='Fixed':
    tracking=False
    tilt_value=df_keys['Tilt/GCR'].loc[df_keys['Project']==project].item()
    surface_azimuth_value=180
    
    #make even fixed tilts a df of values so every time slot has a value and tz localize works the same for fixed vs trackers
    tilt_data=df_data.copy()
    tilt_data['surface_tilt']=tilt_value
    tilt_data['surface_azimuth']=surface_azimuth_value
    surface_azimuth=tilt_data.surface_azimuth
    tilt=tilt_data.surface_tilt
    gcr=None
elif racking =='Tracker':
    tracking=True
    gcr=df_keys['Tilt/GCR'].loc[df_keys['Project']==project].item()
    
    tracker_data=simulate_trackers(project, df_data.index[0], df_data.index[-1], res="5Min")
    surface_azimuth=tracker_data.surface_azimuth
    tilt=tracker_data.surface_tilt
    #tilt=False
    
elevation=df_keys['Elevation'].loc[df_keys['Project']==project].item()
max_angle=df_keys['Max_angle'].loc[df_keys['Project']==project].item()
temp_coef=df_keys['Temp_Coeff_Pmax'].loc[df_keys['Project']==project].item()


        
axis_tilt=0  #


ghi,dni,dhi,airmass,dni_et,site_location,solar_position=decompose(df_data['25.005.BLK01.MET02.PYR02 - GHI.IRRADIANCE_GHI (W/m^2)'],lat,lon,elevation,tz) 

#surface_tilt=20


GHI=df_data['25.005.BLK01.MET02.PYR02 - GHI.IRRADIANCE_GHI (W/m^2)']
albedo=1

asdf
#remove all TZ stuff? I think yes and then put it back to line up with dashboard
tilt=tilt.tz_localize(None)
surface_azimuth=surface_azimuth.tz_localize(None)
solar_position=solar_position.tz_localize(None)
GHI=GHI.tz_localize(None)
dni=dni.tz_localize(None)
dhi=dhi.tz_localize(None)
dni_et=dni_et.tz_localize(None)
airmass=airmass.tz_localize(None)


entries=[tilt, surface_azimuth, solar_position, GHI, dni, dhi, dni_et, airmass]
for entry in entries:
    print str(entry)
    entry.index=half_interval_up(entry.index)

poa_df=transpose(tilt, surface_azimuth, albedo, solar_position, GHI, dni, dhi, dni_et, airmass) 

entries+=[poa_df]
for entry in entries:
    print str(entry)
    entry.index=half_interval_down(entry.index)

POA=poa_df['poa_global']


#plot for checks
plt.figure()
GHI.plot()
tilt.plot()
POA.plot()
surface_azimuth.plot()


plt.figure()
df_data[poa_list].plot()

POA.tz_localize(tz).plot(style='--')


POA_return=POA.resample('H').mean()
POA_return.index=half_interval_up(POA_return.index)
POA_return.tz_localize(tz).plot(style='-.')
POA_return=POA_return.resample('H').mean()
#may have to mess with TZ BS here :( just not sure how it'll line up to dashboard in config sensor stuff
POA_return=POA.tz_localize(None)
POA_return=POA_return.resample('H').mean()


"""

if __name__ == '__main__':
    #project='Wheatland 2B'
    project = 'Holabird Ave 3 Solar'
    start='2022-9-1'
    end='2022-10-1'
    tz='US/'+str(df_keys['Timezone'].loc[df_keys['Project']==project].item())

    df_cats,df_units=solcats_to_dash(project,start,end,resample=True)


    ####### If list index error for GHI, comment out lines 517, 520, and 522. #######

    df_poa=site_s3_to_poa(project,start,end)    # If Solcast data doesn't exist and you ONLY WANT TRANSPOSED GHI, only run this fuction call here
    ##
    ###    timezones will be a pain if you plot on top of eachother/on top of dashboard
    df_cats['sites_ghi_poa']=df_poa
    df_cats[['poa','ghi']].tz_localize(None).plot() #will plot on top of dashboard
    df_cats['sites_ghi_poa'].tz_localize(None).plot()

    df_cats.to_clipboard()




# -*- coding: utf-8 -*-
"""
Created on 12/12/17

-v1: normal meter correction at POI based on PVsyst data


@author: RB + KA
"""


import pandas as pd
import numpy as np

def list_pairs(lis):
    #find pairs that are at least 2 hours in duration
    beg = []
    end = []
    
    start_time = 0
    
    while start_time < len(lis):
        end_time = start_time+1
        while end_time < len(lis) and lis[end_time] - lis[end_time-1] == pd.Timedelta('1h'):
            end_time = end_time+1
        # end_time overshoots by one hour, so compare to > 2 to make sure it's longer than 1 hr
        if end_time - start_time > 2:
            beg.append(lis[start_time])
            end.append(lis[end_time-1])
        start_time = end_time
    return beg, end
    
def Correct_POI_data_v01(df, pos_Meter, pos_Meter_Cum, df_Pvsyst, clipping_KW, df_sensor_ON):
    
    # prevent new columns from polluting the main script
    df = df.copy()
    
    #error handling 
    error_log = pd.DataFrame(columns = ['Site_name', 'year', 'ii', 'p', 'p_max', 'hours'])

    #if no cum meter data, don't go through function
    if df[pos_Meter_Cum].sum().sum() == 0:
        aux_Meter_Corrected = df[pos_Meter].rename(columns = {x:x + '_OK' for x in pos_Meter})
        return aux_Meter_Corrected, [], error_log
    
    #
    aux = pd.DataFrame()
    aux = df[pos_Meter + pos_Meter_Cum]
    #
    #  find if all sensors are ON
    #
    
    #
    #find MEter ON
    #
    
    aux.loc[aux.index, 'ALL_METER_ON'] = aux[pos_Meter].prod(axis=1).values
    aux.loc[aux['ALL_METER_ON']>0,'ALL_METER_ON'] = 1
    aux.loc[aux['ALL_METER_ON']<1,'ALL_METER_ON'] = 0
    #
    '''
    aux['ALL_METER_CUM_ON'] = aux[pos_Meter_Cum].prod(axis=1).values
    aux.loc[aux['ALL_METER_CUM_ON']>0,'ALL_METER_CUM_ON'] = 1
    aux.loc[aux['ALL_METER_CUM_ON']<1,'ALL_METER_CUM_ON'] = 0
    '''
    aux_Meter_Corrected = pd.DataFrame()
    
    draker_flags = []
    #
    #       Loop the meters
    #
    for ii in range(0,len(pos_Meter)):
        
        meter = pos_Meter[ii]
        meter_c_str = pos_Meter_Cum[0].rsplit("_",1)[0]+"_"
        meter_c = meter_c_str + pos_Meter[ii].split("_")[-1] #pos_Meter_Cum[ii]
        #
        aux_m = pd.DataFrame()
        aux_m = aux[[meter] +[meter_c]]
        
        #cum meter can't drop go downwards
        #aux_m.loc[aux_m[meter_c].diff() < 0, [meter] +[ meter_c]] = 0
        aux_m.loc[aux_m[meter_c].diff() < clipping_KW * -0.02, [meter] +[ meter_c]] = 0
        
        # kill big spikes
        #aux_m.loc[df_sensor_ON[meter] == "Over nameplate threshold", meter] = 0
        #aux_m.loc[df_sensor_ON[meter] == "Below negative 10% threshold", meter] = 0
        aux_m.loc[df_sensor_ON[meter] != True, meter] = 0
        
        #find draker hidden outtage
        #aux_flags = aux_m.loc[(aux_m[meter_c].diff() > (1.3 * clipping_KW) / np.sqrt(len(pos_Meter))), :].index     #find spikes above max power
        aux_flags = aux_m.loc[(aux_m[meter_c].diff() > (1.3 * clipping_KW) / len(pos_Meter)), :].index     #find spikes above max power
        aux_flags_2 = aux_m.loc[(df_sensor_ON[meter] == True) & (aux_m[meter] == 0) & (aux_m[meter_c] != 0)].index  #find long lulls of no production
        
        targets = []
        for j in [aux_m.index.get_loc(x) for x in aux_flags]:
            l = 1
            while (aux_m.iloc[j-l].name in aux_flags) or (aux_m.iloc[j-l].name in aux_flags_2):
                targets = targets + [aux_m.iloc[j-l].name]
                l = l + 1
        
        #aux_m.loc[aux_flags, 'spikes'] = 1
        #aux_m.loc[aux_flags_2, 'off'] = 1
        #aux_m.loc[targets, 'targets'] = 1
        #aux_m.to_clipboard()
        
        aux_m.loc[targets, [meter] +[meter_c]] = 0
        draker_flags = draker_flags + targets
        
        
        aux_m = aux_m.fillna(0)
        aux_m['M_CUM_OFF'] = 0
        aux_m.loc[(aux_m[meter_c] == 0), 'M_CUM_OFF'] = 1 #aux_m.loc[(aux_m[meter_c] == 0) & (aux['ALL_INV_CUM_ON'] == 0), 'M_CUM_OFF'] = 1    #exclude times when all inverters are on
        
        aux_m_i = aux_m.loc[aux_m['M_CUM_OFF'] == 1,'M_CUM_OFF']
        aux_m['M_CUM_OFF_1'] = 0                                                        #create blank column first. It wonn't be created below if no outages
        
        #
        for j in range (0, len(aux_m_i)):
            aux_m.loc[aux_m.index == aux_m_i.index[j] ,'M_CUM_OFF_1'] = 1
            aux_m.loc[aux_m.index == aux_m_i.index[j] +  pd.DateOffset(hours = 1),'M_CUM_OFF_1'] = 1
            aux_m.loc[aux_m.index == aux_m_i.index[j] + pd.DateOffset(hours = -1) ,'M_CUM_OFF_1'] = 1
        # find increments in cum meter
        aux_m = aux_m.fillna(0)
        aux_m['M_CUM_OFF_2'] = aux_m['M_CUM_OFF_1'] - aux_m['M_CUM_OFF']
        
        # filter out 1-hour data rows in an outage
        # this prevents the start of one outage being the same as the end of the previous, which confuses the script.
        # if 'M_CUM_OFF' in this row is 0, and the previous row is 1, and the next row is 1, reject it
        aux_curr = aux_m['M_CUM_OFF']
        aux_prev = aux_m['M_CUM_OFF'].shift(1)
        aux_next = aux_m['M_CUM_OFF'].shift(-1)
        bad_indices = list((aux_curr == 0) & (aux_prev == 1) & (aux_next == 1))
        if sum(bad_indices) > 0:
            print("***", meter, ": Removed", sum(bad_indices), "1-hour rows in the middle of an outage")
        aux_m.loc[bad_indices, 'M_CUM_OFF_2'] = 0

        #
        #aux_m_column_name =  ["Missing_" + meter_c]
        #aux_m_Missing = aux_m.loc[aux_m['M_CUM_OFF_2'] == 1, meter_c].diff()[1::2]
        aux_m_Missing = pd.Series(0,index = aux_m.index, name = meter_c, dtype = 'float')
        #
        #  #  Meter Expected Production.  Based on historical data.
        #
        aux_meter_exp = aux.loc[ aux['ALL_METER_ON'] == 1, [meter, meter_c] ]  
        aux_meter_exp['hour'] = aux_meter_exp.index.hour
        #
        df['Hour'] = df.index.hour
        df['Month'] = df.index.month
        #
        #aux_meter_exp = aux_meter_exp.groupby( aux_meter_exp['hour'].values).mean()
        
        #  Meter Expected Production.  Based on PVSyst model data.
        df_Pvsyst['Month'] = df_Pvsyst.index.month
        df_Pvsyst['Hour'] = df_Pvsyst.index.hour
        
        aux_meter_exp = df_Pvsyst.groupby(['Month','Hour'], as_index=False)['POI Output (kWh)'].mean()
        
        # 
        '''
        aux_meter_exp_2 = pd.DataFrame(index = range(0,24))
        aux_meter_exp = pd.concat([aux_meter_exp_2, aux_meter_exp], axis = 1).fillna(0) 
        aux_meter_exp.columns = ['KW_M_Expect', 'KWh_M_Expect','hour']

        aux_meter_exp = aux_meter_exp['KW_M_Expect']
        #
        
        fill_m = pd.DataFrame()
        fill_m[aux_m.columns] = aux_m[aux_m['M_CUM_OFF'] == 1]
        fill_m['hour'] = fill_m.index.hour
        #
        #
        aux_fill = []
        date_fill = []
        for i in range(0, len( fill_m.index)):
            hour_fill = fill_m.index.hour[i]
            aux_fill.append(aux_meter_exp[aux_meter_exp.index == hour_fill].values[0])
            date_fill.append( fill_m.index[i])
        '''
        df['ind'] = df.index
        aux_fill = pd.merge(df[['ind', 'Month','Hour']], aux_meter_exp,  how = 'inner', on=['Month','Hour']).set_index('ind', drop=True)
        aux_fill = aux_fill.loc[df.index.tolist()]
        aux_fill = aux_fill[['POI Output (kWh)']]
        aux_fill.columns = ['E_sum_Fill']
        
        #    
        #aux_fill = pd.DataFrame(aux_fill, index = date_fill, columns = ['E_sum_Fill'] )
        #aux_fill = pd.DataFrame(aux_fill, index = date_fill)                                   #changed by ryan due to crashing when no meter outages
        #aux_fill = aux_fill.set_index(aux_m[aux_m['M_CUM_OFF'] == 1].index)
        #aux_fill.columns = ['E_sum_Fill']                                                      #changed by ryan due to crashing
        #
        aux_f = pd.concat([aux,aux_m, aux_fill], axis=1)
        aux_f = aux_f.fillna(0)
        aux_f.rename(columns={0:'Missing_M'}, inplace=True)
        #
        #  Distribute energy
        pos = aux_f.loc[aux_f['M_CUM_OFF_2'] > 0,  'M_CUM_OFF_2'].index 
        #
        
        # if the file starts out with missing data, then the end of this block will mistakenly be
        # identified as the beginning of a block.  Check to see if this is the case, and correct if necessary
        if aux_f['M_CUM_OFF'][0] == 1:
            pos = pos[1:]
        # similarly, if the file ends with missing data, remove the last element of pos
        if aux_f['M_CUM_OFF'][-1] == 1:
            pos = pos[:-1]
        
        if (len(pos) % 2) == 1: # "odd"
            print("***", meter, ": Number of (start, end) outage pairs is mismatched")
        #        
        #  Only loop if Energy is bigger than 300 W
        aux_m_Missing = aux_m.loc[pos, meter_c].diff()[1::2]
        aux_m_Missing = aux_m_Missing[aux_m_Missing > 300]
        aux_m_Missing = aux_m_Missing.rename('Missing_KW')
        aux_f = pd.concat([aux_f, aux_m_Missing], axis=1)


        aux_f['Missing_M_ADDED'] = 0               #initial column to fill data
        
        
        for p in range(0,len(pos),2): 
            aux_e_filled_value = []                 #for each loop, blank data and start over
            aux_e_filled_date = []
            
            e_to_distribute = aux_f['E_sum_Fill'][pos[p]:pos[p+1]]
            e_to_distribute = e_to_distribute.sort_values( ascending = False)
            e_to_fill = aux_f['Missing_KW'][pos[p+1]] 
            #print p, e_to_fill
            #
            k = 0
            while k < len(e_to_distribute):
                if e_to_distribute[k] < e_to_fill:
                    aux_e_filled_value.append(e_to_distribute[k])
                    aux_e_filled_date.append(e_to_distribute.index[k])
                    e_to_fill = e_to_fill - e_to_distribute[k] 
                #else:
                    #k = len(e_to_distribute)               
                k = k+1
            # 
                
            ##input scale factor due to left over energy            
            if aux_e_filled_value != []:
                aux_factor = (np.sum(aux_e_filled_value) +e_to_fill) / np.sum(aux_e_filled_value)
                if aux_factor < 1.5:
                    aux_e_filled_value = [x * aux_factor for x in aux_e_filled_value]       #multiply factor through
                else:
                    aux_factor = 1.5                
                    aux_e_filled_value = [x * aux_factor for x in aux_e_filled_value]       #multiply factor through
                aux_e_filled_value = [min(x, clipping_KW) for x in aux_e_filled_value]
            
        
            #create export dataframe
            aux_e_filled = pd.DataFrame(aux_e_filled_value, index = aux_e_filled_date)
            aux_e_filled.rename(columns={0:'Missing_M_ADDED'}, inplace=True) 
            #export out of loop
            if not aux_e_filled.empty:              #if it has data, append it in. If empty just skip
                aux_f.loc[ (aux_e_filled.index), 'Missing_M_ADDED'] = aux_e_filled['Missing_M_ADDED'].values
                
        #
        aux_f = aux_f.fillna(0)
        aux_f['Missing_M_ADDED'] = aux_f['Missing_M_ADDED'].multiply(aux_f['M_CUM_OFF'])
        name_c = meter +'_OK'
        #
        aux_Meter_Corrected[name_c] = aux_m[meter] + aux_f['Missing_M_ADDED']  #combined corrected sensor
        #aux_Meter_Corrected.loc[ aux_Meter_Corrected[name_c] < 0, name_c] = 0      #remove off hours
        
        
        ## Check work using cumulative sum
        # remove day time points
        aux_ind = aux_f.loc[(aux_f.index.hour <= 4) | (aux_f.index.hour >= 22), :]
        aux_f['Check'] = 0                          #create identifier column if WRAPPED around loop range
        aux_f['Pos'] = 0                            #create identifier column if within loop range

        
        #for each loop in pos
        for p in range(0,len(pos),2):
            #ignore times when meter starts or ends in outage (causes error and no corrections anyway)
            if pos[p] != aux_f.index[0] and pos[p+1] != aux_f.index[-1]:
                aux_f.loc[(aux_f.index <= pos[p+1]) & (pos[p] <= aux_f.index), 'Pos'] = 1
                
                
                try:
                    check_beg = aux_ind.loc[(aux_ind.index < pos[p]) & (aux_ind['M_CUM_OFF'] == 0), :].index[-1]
                    check_end = aux_ind.loc[(pos[p+1] < aux_ind.index) & (aux_ind['M_CUM_OFF'] == 0), :].index[0]
                    aux_f.loc[(check_beg <= aux_f.index) & (aux_f.index <= check_end), 'Check' ] = 1
                except:
                    error_append = pd.DataFrame(index = [0], columns = error_log.columns.tolist())
                    error_append[[u'ii', u'p', 'p_max', u'hours']] = [ii, p, range(0,len(pos),2)[-1], (pos[p+1] - pos[p]).seconds/60/60]
                    error_log = error_log.append(error_append)
                    pass

                
                
        
        #find outer 
        beg_list, end_list = list_pairs( aux_f.loc[aux_f['Check'] == 1, :].index.tolist() )
        
        #loop through outer check and adjust energy
        for x in range(len(beg_list)):
            start = beg_list[x]
            finish = end_list[x]
            
            cum_energy = aux_m.loc[finish, meter_c].item() - aux_m.loc[start, meter_c].item()
            meter_total = aux_Meter_Corrected.loc[(start <= aux_f.index) & (aux_f.index <= finish), name_c].sum()
            meter_added = aux_Meter_Corrected.loc[(start <= aux_f.index) & (aux_f.index <= finish) & (aux_f['Pos'] == 1), name_c]
            meter_sum = meter_added.sum()
            
            delta = cum_energy - meter_total            
            
            if meter_sum > 0:
                second_factor = (meter_sum + delta) / meter_sum
            else:
                #can't make any adjustments if no energy was added
                second_factor = 1
                
            if (second_factor < 1.05) and (second_factor > 0.1) and (cum_energy > 0):
                meter_adjusted = meter_added * second_factor
                aux_Meter_Corrected.loc[meter_adjusted.index, name_c] = meter_adjusted
                
            #also zero out any shoulder times
            #if (second_factor < 1.05) and (second_factor < 0) and (cum_energy > 0):
            #    meter_adjusted = meter_added * 0.00001
            #    aux_Meter_Corrected.loc[meter_adjusted.index, name_c] = meter_adjusted                
            
            
            
            
        
        
        
    return aux_Meter_Corrected, draker_flags, error_log


# -*- coding: utf-8 -*-
"""
Created on Wed Jun 22 14:16:25 2016

@author: SEBASTIAN

Change log:

- 2016/07/24 -> Added code to make sure that days 1/1 , 7/4 and 12/25 are holidays.  
- 2017/07/01 -> Fixed DST not working. DST start and end were read from config instead of being calculated
- 2017/07/13 -> Speed up code to run on peak hours for energy and capacity. 
- 2017/08/09 -> Add line-loss adjustment
TODO: read holiday list from config file

"""

import pandas as pd

import numpy as np
import datetime as dt
import os
from holidays import UnitedStates
import calendar


from pandas.tseries.holiday import AbstractHolidayCalendar, Holiday, nearest_workday, \
    USMartinLutherKingJr, USPresidentsDay, GoodFriday, USMemorialDay, \
    USLaborDay, USThanksgivingDay, TH, FR, SA, next_monday

class USTradingCalendar(AbstractHolidayCalendar):
    rules = [
        Holiday('NewYearsDay', month=1, day=1, observance=nearest_workday),
        #USMartinLutherKingJr,
        #USPresidentsDay,
        GoodFriday,
        USMemorialDay,
        Holiday('USIndependenceDay', month=7, day=4, observance=nearest_workday),
        USLaborDay,
        USThanksgivingDay,      
        Holiday('Christmas', month=12, day=25, observance=nearest_workday)
    ]

def unpivot(frame):
    frame_cols = frame.columns.tolist()
    output = pd.melt(frame, id_vars = frame_cols[0], value_vars = frame_cols[1:])
    return output

def get_trading_close_holidays(year):
    inst = USTradingCalendar()
    return inst.holidays(dt.datetime(year-1, 12, 31), dt.datetime(year, 12, 31))

def generate_Holidays(year):
    dr = pd.date_range(dt.datetime(year, 1, 1), dt.datetime(year+1, 1, 1), freq='H')    
    df = pd.DataFrame()
    df['Date'] = dr
    holidays = get_trading_close_holidays(year)
    # add day after Thanksgiving    
    Black_Friday = holidays[5] + dt.timedelta(days=1)    
    l_holidays = []
    #
    for i in range(0,len(holidays)):
        for j in range(0,24):
            l_holidays.append(holidays[i]  + dt.timedelta(hours = j) )
    #  add Black_Friday
    #
    for j in range(0,24):
            l_holidays.append(Black_Friday  + dt.timedelta(hours = j) )        
            
    df['Holiday'] = df['Date'].isin(l_holidays)    
    
    #df.set_value(idx, 'Holiday', True)
    df = df.set_index('Date')
	#
    #  Make sure that these days are holidays
    df.loc[(df.index.month == 1) & (df.index.day == 1), 'Holiday'] = True
    df.loc[(df.index.month == 7) & (df.index.day == 4), 'Holiday'] = True
    df.loc[(df.index.month == 12) & (df.index.day == 25), 'Holiday'] = True           

    return df

def yearly8760(file_input, shift_DST, year):
    #read new file - only use specific year's worth of data
    path, config = os.path.split(file_input)
    project = file_input.split(os.sep)[-3]
    read_file = os.path.join(path, project + '_YearlyRates.csv')
    df = pd.read_csv(read_file, index_col = 'IND', usecols= ['IND', np.str(year)])
    
    #initialize dataframe
    df.rename(columns = {np.str(year): 'Rates' }, inplace= True)
    try:
        Flat_Capacity = df.loc['Flat_Capacity',:].values[0]
        df.drop('Flat_Capacity', inplace = True)
    except:
        Flat_Capacity = 0
    Flat_Total = Flat_Capacity + 0      #sum all yearly flat charges here: capacity, fees, etc
    
    #fix time index
    dr = pd.date_range(dt.datetime(year, 1, 1), dt.datetime(year+1, 1, 1), freq='H')
    if calendar.isleap(year):
        dr = dr[~((dr.month == 2) & (dr.day == 29))]
    if ((len(dr) == 8761) & (len(df) == 8760)):
        df.loc['add last_row',:] = 0
        df.set_index(dr, inplace = True)
    else:
        'something wrong with yearly rate calc, give up'
        print(project + ' Yearly Rate Calc Failed***')
        df = pd.DataFrame(np.zeros(len(dr)), index=dr, columns = ['Rates'])
        
    #Reverse DST shift - True for Pvsyst, False for AE data. Rate calc is already DST shifted as is the AE data. Must reverse DST shift for pvsyst
    if shift_DST:
        #second sunday in March
        d = dt.datetime(year, 3, 1)
        RATE_dst_start = d+ dt.timedelta(days= 7 + 7 - d.weekday() - 1)   
        #first sunday in November
        x = dt.datetime(year,11,1)
        RATE_dst_end = x + dt.timedelta(days= 7 - x.weekday() - 1)
        
        aux = pd.DataFrame([], index = df.index)
        aux['DST'] = 0
        aux.loc[((aux.index > RATE_dst_start) & (aux.index < RATE_dst_end)),:] = 1
        aux_2 = df[aux['DST'] == 1].shift(-1).fillna(0)
        df.loc[aux_2.index,'Rates'] = aux_2['Rates']
        
    #Fabricate leap day for leap years
    if calendar.isleap(year):
        dlp = pd.date_range(dt.datetime(year, 2, 29), dt.datetime(year, 3, 1), freq='H')[:-1]
        
        #get all the prices for the same day of the week
        aux = df.loc[(df.index.month ==2) & (df.index.dayofweek == dlp.dayofweek[0])]
        aux['hour'] = aux.index.hour
        leap_rates = aux.groupby('hour').mean().values
        
        d_add = pd.DataFrame(data = leap_rates, index = dlp, columns = ['Rates'])
        
        df = df.append(d_add)
        df = df.sort_index(ascending = True)
    
    df['Rates'] = df['Rates'].astype(float)
    df['Flat'] = Flat_Total / len(df.index)
    filler = ['Holiday','month','DST','weekday','Peak_day','ON_Peak','Summer','Energy_Peak','Capacity_Peak']
    
    aux = pd.DataFrame(np.zeros([len(df),len(filler)]), index = df.index, columns = filler)
    
    export = pd.concat([aux,df], axis = 1)
    
    return export

def generate_Rate_column (file_input, shift_DST, year ):
    #
    #shift_DST = True to calculate values from 8760.  Otherwise is False
   
    ##added 7/8/21 to deal with values column coming in as NaN
#    import win32com.client
#    import time
#    
#    xl = win32com.client.Dispatch("Excel.Application")
#    wb = xl.Workbooks.Open(file_input)
#    wb.RefreshAll()
#    time.sleep(5)
#    wb.Save()
#    wb.Close(True)
#    xl.Quit()
#    
    try:
        df_t1 = pd.read_excel(file_input, sheet_name = 'Table_1')
        print('got df_t1')
        df_t2 = pd.read_excel(file_input, sheet_name = 'Table_2', index_col=0).fillna(0)
        normal_rate_flag = True
    except:
        #rate calc not in config file, redirect to yearly rates file
        normal_rate_flag = False
        return yearly8760(file_input, shift_DST, year), normal_rate_flag

        
    #  Create DataFrame with Columns of Interest
    #
    #   Read all variables defined on Table 1
    #
    # year =  df_t1[df_t1['Name'] == 'Current Year']['Value'].values[0]
    #fixed 6/27/17, reads raw data instead of calculating DST dates
    #    RATE_dst_start = df_t1[df_t1['Name'] == 'DST_start']['Value'].values[0]
    #    RATE_dst_end = df_t1[df_t1['Name'] == 'DST_end']['Value'].values[0]
    d = dt.datetime(year, 3, 1)
    RATE_dst_start = d + dt.timedelta(days= 7 + 7 - d.weekday() - 1)
    
    x = dt.datetime(year,11,1)
    RATE_dst_end = x + dt.timedelta(days= 7 - x.weekday() - 1)
    
    On_Peak_Sunday    =  df_t1[df_t1['Name'] == 'On_Peak_Sunday']['Value'].values[0]
    On_Peak_Monday    =  df_t1[df_t1['Name'] == 'On_Peak_Monday']['Value'].values[0]   
    On_Peak_Tuesday   =  df_t1[df_t1['Name'] == 'On_Peak_Tuesday']['Value'].values[0]
    On_Peak_Wednesday =  df_t1[df_t1['Name'] == 'On_Peak_Wednesday']['Value'].values[0]
    On_Peak_Thursday  =  df_t1[df_t1['Name'] == 'On_Peak_Thursday']['Value'].values[0]
    On_Peak_Friday    =  df_t1[df_t1['Name'] == 'On_Peak_Friday']['Value'].values[0]
    On_Peak_Saturday  =  df_t1[df_t1['Name'] == 'On_Peak_Saturday']['Value'].values[0]
    
    Cap_Summer_1  =  df_t1[df_t1['Name'] == 'Cap_Summer_1']['Value'].values[0]
    Cap_Summer_2  =  df_t1[df_t1['Name'] == 'Cap_Summer_2']['Value'].values[0]
    Cap_Summer_3  =  df_t1[df_t1['Name'] == 'Cap_Summer_3']['Value'].values[0]
    Cap_Summer_4  =  df_t1[df_t1['Name'] == 'Cap_Summer_4']['Value'].values[0]
    Cap_Summer_5  =  df_t1[df_t1['Name'] == 'Cap_Summer_5']['Value'].values[0]
    Cap_Summer_6  =  df_t1[df_t1['Name'] == 'Cap_Summer_6']['Value'].values[0]
    Cap_Summer_7  =  df_t1[df_t1['Name'] == 'Cap_Summer_7']['Value'].values[0]
    Cap_Summer_8  =  df_t1[df_t1['Name'] == 'Cap_Summer_8']['Value'].values[0]
    Cap_Summer_9  =  df_t1[df_t1['Name'] == 'Cap_Summer_9']['Value'].values[0]
    Cap_Summer_10  =  df_t1[df_t1['Name'] == 'Cap_Summer_10']['Value'].values[0]
    Cap_Summer_11  =  df_t1[df_t1['Name'] == 'Cap_Summer_11']['Value'].values[0]
    Cap_Summer_12  =  df_t1[df_t1['Name'] == 'Cap_Summer_12']['Value'].values[0]
    
    LINE_LOSS = df_t1[df_t1['Name'] == 'Line Loss  Rate Increase (applies to ENERGY rates only)']['Value'].values[0]
        
    ENERGY_ON_PEAK  =  (1+LINE_LOSS) * df_t1[df_t1['Name'] == 'ENERGY_ON_PEAK_[$/KWh]']['Value'].values[0]
    ENERGY_OFF_PEAK =  (1+LINE_LOSS) * df_t1[df_t1['Name'] == 'ENERGY_OFF_PEAK_[$/KWh]']['Value'].values[0]
    
    CAPACITY_ON_PEAK_SUMMER =  df_t1[df_t1['Name'] == 'CAPACITY_ON_PEAK_SUMMER_[$/KWh]']['Value'].values[0]
    CAPACITY_ON_PEAK_NON_SUMMER =  df_t1[df_t1['Name'] == 'CAPACITY_ON_PEAK_NON_SUMMER_[$/KWh]']['Value'].values[0]
    
    
    #---------------
    #  Creates Index with Dates and Column with Holidays info
    df = generate_Holidays(year)
    #  
    df['month'] = df.index.month
    df['hour'] = df.index.hour
    #
    #---------------
    #  DST?
    #
    #    aux = pd.DataFrame(np.zeros(shape=(len(df), 1)), columns = ['DST'])
    #    aux = aux.set_index(df.index)
    #    aux_2 = aux[RATE_dst_start : RATE_dst_end] +1
    #    df['DST'] = aux.add(aux_2, fill_value = 0)
    df['DST'] = 0
    df.loc[(df.index > RATE_dst_start) & (df.index < RATE_dst_end), 'DST'] = 1
    
    #---------------
    #  Weekday
    df['weekday'] = df.index.weekday
    #
    #  Find holidays that are in Sundays
    #holidays_and_Sunday = df[df['weekday'] == 6 & df['holiday'] == ]
    #---------------
    #  Peak Day
    #  Monday = 0,..., Sunday = 6  (In excel is Sunday is 1)
    df['Peak_day'] = 0
    
    df.loc[df['weekday'] == 6, ['Peak_day']] = On_Peak_Sunday
    df.loc[df['weekday'] == 0, ['Peak_day']] = On_Peak_Monday
    df.loc[df['weekday'] == 1, ['Peak_day']] = On_Peak_Tuesday
    df.loc[df['weekday'] == 2, ['Peak_day']] = On_Peak_Wednesday
    df.loc[df['weekday'] == 3, ['Peak_day']] = On_Peak_Thursday
    df.loc[df['weekday'] == 4, ['Peak_day']] = On_Peak_Friday
    df.loc[df['weekday'] == 5, ['Peak_day']] = On_Peak_Saturday
    #
    df.loc[df['Holiday'] == True, ['Peak_day']] = 0
    
    #---------------
    #  On Peak  [AND( peak day, NOT Holiday)]
    
    df['ON_Peak'] = 0
    df.loc[(df['Peak_day'] == 1) & (df['Holiday'] == False), ['ON_Peak']] = 1
    
    #  Summer (1) or Not Summer (0)
    df['Summer'] = 0
    
    df.loc[df['month'] == 1, ['Summer']] = Cap_Summer_1
    df.loc[df['month'] == 2, ['Summer']] = Cap_Summer_2
    df.loc[df['month'] == 3, ['Summer']] = Cap_Summer_3
    df.loc[df['month'] == 4, ['Summer']] = Cap_Summer_4
    df.loc[df['month'] == 5, ['Summer']] = Cap_Summer_5
    df.loc[df['month'] == 6, ['Summer']] = Cap_Summer_6
    df.loc[df['month'] == 7, ['Summer']] = Cap_Summer_7
    df.loc[df['month'] == 8, ['Summer']] = Cap_Summer_8
    df.loc[df['month'] == 9, ['Summer']] = Cap_Summer_9
    df.loc[df['month'] == 10, ['Summer']] = Cap_Summer_10
    df.loc[df['month'] == 11, ['Summer']] = Cap_Summer_11
    df.loc[df['month'] == 12, ['Summer']] = Cap_Summer_12
    #    
    #---------------
    #  Energy Peak 
    #
    #  Extract table from Table_2 sheet.
    ep_t = df_t2.iloc[2:26].astype(int)
     # shift hours from 1-24 to 0-23
    ep_t['ENERGY PEAK HOURS'] -= 1
    #prep for unpivot and melt data
    ep_t.set_index(ep_t[ep_t.columns[0]], drop = True, inplace = True)
    ep_t.columns = range(0,13)
    frame = unpivot(ep_t)
    
    #commute peak hours back to df
    frame['string'] = frame[[0, u'variable']].astype(str).apply(lambda x: '--'.join(x), axis=1)
    lis = frame.loc[ frame['value'] == 1, 'string'].tolist()

    aux_df = pd.DataFrame([], df.index)
    aux_df['string'] = df[['hour', 'month']].astype(str).apply(lambda x: '--'.join(x), axis=1)
    
    df['Energy_Peak'] = 0
    df.loc[ aux_df['string'].isin(lis), 'Energy_Peak'] = 1
    
    aux_2 = []
       
          #  Multiply by Colun On Peak
    df['Energy_Peak'] = df['Energy_Peak'].multiply(df['ON_Peak'], axis = 'index')
    
    #  Shift Values in Energy Peak Based on DST Column
    
    if shift_DST == True:
        aux = df['Energy_Peak']    
        aux = aux [df['DST'] == 1].shift(-1).fillna(0)
        aux_2 =   pd.concat([df['Energy_Peak'] , aux])
        aux_2 = aux_2[~aux_2.index.duplicated(keep='last')]
        df['Energy_Peak'] = aux_2
    
    #---------------
    #  CAPACITY Peak? 
    #
    #  Extract table from Table_2 sheet.
    ep_t = df_t2.iloc[29:54].astype(int)
    # shift hours from 1-24 to 0-23
    ep_t['ENERGY PEAK HOURS'] -= 1
  
    ep_t.set_index(ep_t[ep_t.columns[0]], drop = True, inplace = True)
    ep_t.columns = range(0,13)
    frame = unpivot(ep_t)
    frame['string'] = frame[[0, u'variable']].astype(str).apply(lambda x: '--'.join(x), axis=1)
    lis = frame.loc[ frame['value'] == 1, 'string'].tolist()

    aux_df = pd.DataFrame([], df.index)
    aux_df['string'] = df[['hour', 'month']].astype(str).apply(lambda x: '--'.join(x), axis=1)
    
    df['Capacity_Peak'] = 0
    df.loc[ aux_df['string'].isin(lis), 'Capacity_Peak'] = 1

       
          #  Multiply by Colun On Peak
    df['Capacity_Peak'] = df['Capacity_Peak'].multiply(df['ON_Peak'], axis = 'index')
        #  Shift Values in Energy Peak Based on DST Column

    if shift_DST == True:
        aux = df['Capacity_Peak']    
        aux = aux [df['DST'] == 1].shift(-1).fillna(0)
        aux_2 =   pd.concat([df['Capacity_Peak'] , aux])
        aux_2 = aux_2[~aux_2.index.duplicated(keep='last')]
        df['Capacity_Peak'] = aux_2   
    #
    #---------------
    
    df['Rates'] = ENERGY_OFF_PEAK
    
    df.loc[(df['ON_Peak'] == 1) & (df['Summer'] == 1) & (df['Capacity_Peak'] == 1), ['Rates']] = ENERGY_ON_PEAK + CAPACITY_ON_PEAK_SUMMER 
    df.loc[(df['ON_Peak'] == 1) & (df['Summer'] == 0) & (df['Capacity_Peak'] == 1), ['Rates']] = ENERGY_ON_PEAK + CAPACITY_ON_PEAK_NON_SUMMER 
    
    df['Flat'] = 0              #unless read from monthly charges
    
    del df['hour']
    
    return df, normal_rate_flag

